/*************************************************
		GlcdPlus.cpp
		Ver. 0.95

	Copyright (C) 2010-2011,
		MicroFan, Div. PeopleMedia, Inc.
	All rights reserved.
*************************************************/


#include "GlcdPlus.h"

#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include "WProgram.h"
#include <avr/pgmspace.h>


/*************************************************
* SpiGpio
*************************************************/

SpiGpio::SpiGpio(uint8_t ss, uint8_t addr)
{
    pinSS = ss ;
    digitalWrite(pinSS, HIGH) ;
    pinMode(pinSS, OUTPUT) ; // ~SS

    pinMode(13, OUTPUT) ; // SCK
    pinMode(12, INPUT) ; // MISO
    pinMode(11, OUTPUT) ; // MOSI

    devAddr = addr ;

#if defined(__AVR_ATmega1280__)
    digitalWrite(13, HIGH) ;
    digitalWrite(11, HIGH) ;
#else
    pinMode(10, OUTPUT) ; // SS
/*
    if (!(SPCR & (1 << SPE))) { // ɐݒ肳ĂȂ
        SPCR = (1 << SPE) | (1 << MSTR) ;
        SPSR |= (1 << SPI2X) ;
    }
*/
#endif
}

uint8_t SpiGpio::access_addr8(uint8_t cmd, uint8_t reg, uint8_t data)
{
#if defined(__AVR_ATmega1280__)
    uint8_t d, so ;
    int i ;

    digitalWrite(pinSS, LOW) ;

    d = cmd | (devAddr << 1) ;
    for (i = 0; i < 8; i++) {
        so = (d & 0x80) ? 0x20:0 ;
        PORTB = (PORTB & 0x5F) | so ; // SCK & MO
        d <<= 1 ;
        PORTB |= 0x80 ; // SCK
    }

    d = reg ;
    for (i = 0; i < 8; i++) {
        so = (d & 0x80) ? 0x20:0 ;
        PORTB = (PORTB & 0x5F) | so ; // SCK & MO
        d <<= 1 ;
        PORTB |= 0x80 ;
    }

    d = data ;
    for (i = 0; i < 8; i++) {
        so = (d & 0x80) ? 0x20:0 ;
        PORTB = (PORTB & 0x5F) | so ; // SCK & MO
        d <<= 1 ;
        PORTB |= 0x80 ;
        d |= (PINB & 0x40) ? 1 : 0 ;
    }

    digitalWrite(pinSS, HIGH) ;

    return d ;
#else
    uint8_t xspcr, xspsr ;

    xspcr = SPCR ;
    xspsr = SPSR ;  
    SPCR = (1 << SPE) | (1 << MSTR) ;
    SPSR |= (1 << SPI2X) ;

    digitalWrite(pinSS, LOW) ;

    SPDR = cmd | (devAddr << 1) ;
    while (!(SPSR & (1 << SPIF)))
        ;
    SPDR = reg ;
    while (!(SPSR & (1 << SPIF)))
        ;
    SPDR = data ;
    while (!(SPSR & (1 << SPIF)))
        ;

    digitalWrite(pinSS, HIGH) ;

    SPCR = xspcr ;
    SPSR = xspsr ;  

    return SPDR ;
#endif
}

inline void
    SpiGpio::write(uint8_t reg, uint8_t data)
{
    access_addr8(GPIO_WRITE, reg, data) ;
}

inline uint8_t
    SpiGpio::read(uint8_t reg)
{
    return access_addr8(GPIO_READ, reg, 0x00) ;
}


/*************************************************
* GlcdPlus
*************************************************/

// #include <avr/pgmspace.h>

const uint8_t FONT[] PROGMEM = {

    /*  !"#$%&'()*+,-./ */

    0b00000000,
    0b00000000,
    0b00000000,
    0b00000000,
    0b00000000,

    0b00000000,
    0b00000000,
    0b01001111,
    0b00000000,
    0b00000000,

    0b00000000,
    0b00000111,
    0b00000000,
    0b00000111,
    0b00000000,

    0b00010100,
    0b01111111,
    0b00010100,
    0b01111111,
    0b00010100,

    0b00100100,
    0b00101010,
    0b01111111,
    0b00101010,
    0b00010010,

    0b00100011,
    0b00010011,
    0b00001000,
    0b01100100,
    0b01100010,

    0b00110110,
    0b01001001,
    0b01010101,
    0b00100010,
    0b01010000,

    0b00000000,
    0b00000101,
    0b00000011,
    0b00000000,
    0b00000000,

    0b00000000,
    0b00011100,
    0b00100010,
    0b01000001,
    0b00000000,

    0b00000000,
    0b01000001,
    0b00100010,
    0b00011100,
    0b00000000,

    0b00010100,
    0b00001000,
    0b00111110,
    0b00001000,
    0b00010100,

    0b00001000,
    0b00001000,
    0b00111110,
    0b00001000,
    0b00001000,

    0b00000000,
    0b01010000,
    0b00110000,
    0b00000000,
    0b00000000,

    0b00001000,
    0b00001000,
    0b00001000,
    0b00001000,
    0b00001000,

    0b00000000,
    0b01100000,
    0b01100000,
    0b00000000,
    0b00000000,

    0b00100000,
    0b00010000,
    0b00001000,
    0b00000100,
    0b00000010,


    /* 0123456789:;<=>? */

    0b00111110,
    0b01010001,
    0b01001001,
    0b01000101,
    0b00111110,

    0b00000000,
    0b01000010,
    0b01111111,
    0b01000000,
    0b00000000,

    0b01000010,
    0b01100001,
    0b01010001,
    0b01001001,
    0b01000110,

    0b00100001,
    0b01000001,
    0b01000101,
    0b01001011,
    0b00110001,

    0b00011000,
    0b00010100,
    0b00010010,
    0b01111111,
    0b00010000,

    0b00100111,
    0b01000101,
    0b01000101,
    0b01000101,
    0b00111001,

    0b00111100,
    0b01001010,
    0b01001001,
    0b01001001,
    0b00110000,

    0b00000001,
    0b01110001,
    0b00001001,
    0b00000101,
    0b00000011,

    0b00110110,
    0b01001001,
    0b01001001,
    0b01001001,
    0b00110110,

    0b00000110,
    0b01001001,
    0b01001001,
    0b00101001,
    0b00011110,

    0b00000000,
    0b00110110,
    0b00110110,
    0b00000000,
    0b00000000,

    0b00000000,
    0b01010110,
    0b00110110,
    0b00000000,
    0b00000000,

    0b00001000,
    0b00010100,
    0b00100010,
    0b01000001,
    0b00000000,

    0b00010100,
    0b00010100,
    0b00010100,
    0b00010100,
    0b00010100,

    0b00000000,
    0b01000001,
    0b00100010,
    0b00010100,
    0b00001000,

    0b00000010,
    0b00000001,
    0b01010001,
    0b00001001,
    0b00000110,


    /* @ABCDEFGHIJKLMNO */

    0b00110010,
    0b01001001,
    0b01111001,
    0b01000001,
    0b00111110,

    0b01111110,
    0b00010001,
    0b00010001,
    0b00010001,
    0b01111110,

    0b01111111,
    0b01001001,
    0b01001001,
    0b01001001,
    0b00110110,

    0b00111110,
    0b01000001,
    0b01000001,
    0b01000001,
    0b00100010,

    0b01111111,
    0b01000001,
    0b01000001,
    0b00100010,
    0b00011100,

    0b01111111,
    0b01001001,
    0b01001001,
    0b01001001,
    0b01000001,

    0b01111111,
    0b00001001,
    0b00001001,
    0b00001001,
    0b00000001,

    0b00111110,
    0b01000001,
    0b01001001,
    0b01001001,
    0b00111010,

    0b01111111,
    0b00001000,
    0b00001000,
    0b00001000,
    0b01111111,

    0b00000000,
    0b01000001,
    0b01111111,
    0b01000001,
    0b00000000,

    0b00100000,
    0b01000000,
    0b01000001,
    0b00111111,
    0b00000001,

    0b01111111,
    0b00001000,
    0b00010100,
    0b00100010,
    0b01000001,

    0b01111111,
    0b01000000,
    0b01000000,
    0b01000000,
    0b01000000,

    0b01111111,
    0b00000010,
    0b00000100,
    0b00000010,
    0b01111111,

    0b01111111,
    0b00000100,
    0b00001000,
    0b00010000,
    0b01111111,

    0b00111110,
    0b01000001,
    0b01000001,
    0b01000001,
    0b00111110,


    /* PQRSTUVWXYZ[\]^_ */

    0b01111111,
    0b00001001,
    0b00001001,
    0b00001001,
    0b00000110,

    0b00111110,
    0b01000001,
    0b01010001,
    0b00100001,
    0b01011110,

    0b01111111,
    0b00001001,
    0b00011001,
    0b00101001,
    0b01000110,

    0b01000110,
    0b01001001,
    0b01001001,
    0b01001001,
    0b00110001,

    0b00000001,
    0b00000001,
    0b01111111,
    0b00000001,
    0b00000001,

    0b00111111,
    0b01000000,
    0b01000000,
    0b01000000,
    0b00111111,

    0b00011111,
    0b00100000,
    0b01000000,
    0b00100000,
    0b00011111,

    0b00111111,
    0b01000000,
    0b00111111,
    0b01000000,
    0b00111111,

    0b01100011,
    0b00010100,
    0b00001000,
    0b00010100,
    0b01100011,

    0b00000111,
    0b00001000,
    0b01110000,
    0b00001000,
    0b00000111,

    0b01100001,
    0b01010001,
    0b01001001,
    0b01000101,
    0b01000011,

    0b00000000,
    0b01111111,
    0b01000001,
    0b01000001,
    0b00000000,

    0b00010101,
    0b00010110,
    0b01111100,
    0b00010110,
    0b00010101,

    0b00000000,
    0b01000001,
    0b01000001,
    0b01111111,
    0b00000000,

    0b00000100,
    0b00000010,
    0b00000001,
    0b00000010,
    0b00000100,

    0b01000000,
    0b01000000,
    0b01000000,
    0b01000000,
    0b01000000,


    /* `abcdefghijklmno */

    0b00000000,
    0b00000001,
    0b00000010,
    0b00000100,
    0b00000000,

    0b00100000,
    0b01010100,
    0b01010100,
    0b01010100,
    0b01111000,

    0b01111111,
    0b01001000,
    0b01000100,
    0b01000100,
    0b00111000,

    0b00111000,
    0b01000100,
    0b01000100,
    0b01000100,
    0b00100000,

    0b00111000,
    0b01000100,
    0b01000100,
    0b01001000,
    0b01111111,

    0b00111000,
    0b01010100,
    0b01010100,
    0b01010100,
    0b00011000,

    0b00001000,
    0b01111110,
    0b00001001,
    0b00000010,
    0b00000000,

    0b00001100,
    0b01010010,
    0b01010010,
    0b01010010,
    0b00111110,

    0b01111111,
    0b00001000,
    0b00000100,
    0b00000100,
    0b01111000,

    0b00000000,
    0b01000100,
    0b01111101,
    0b01000000,
    0b00000000,

    0b00100000,
    0b01000000,
    0b01000100,
    0b00111101,
    0b00000000,

    0b01111111,
    0b00010000,
    0b00101000,
    0b01000100,
    0b00000000,

    0b00000000,
    0b01000001,
    0b01111111,
    0b01000000,
    0b00000000,

    0b01111100,
    0b00000100,
    0b01111000,
    0b00000100,
    0b01111000,

    0b01111100,
    0b00001000,
    0b00000100,
    0b00000100,
    0b01111000,

    0b00111000,
    0b01000100,
    0b01000100,
    0b01000100,
    0b00111000,

    /* pqrstuvwxyz<|>__ */

    0b01111100,
    0b00010100,
    0b00010100,
    0b00010100,
    0b00001000,

    0b00011000,
    0b00100100,
    0b00100100,
    0b00011000,
    0b01111100,

    0b01111100,
    0b00001000,
    0b00000100,
    0b00000100,
    0b00001000,

    0b01001000,
    0b01010100,
    0b01010100,
    0b01010100,
    0b00100000,

    0b00000100,
    0b00111111,
    0b01000100,
    0b01000000,
    0b00100000,

    0b00111100,
    0b01000000,
    0b01000000,
    0b00100000,
    0b01111100,

    0b00011100,
    0b00100000,
    0b01000000,
    0b00100000,
    0b00011100,

    0b00111100,
    0b01000000,
    0b00111100,
    0b01000000,
    0b00111100,

    0b01000100,
    0b00101000,
    0b00010000,
    0b00101000,
    0b01000100,

    0b00001100,
    0b01010000,
    0b01010000,
    0b01010000,
    0b00111100,

    0b01000100,
    0b01100100,
    0b01010100,
    0b01001100,
    0b01000100,

    0b00000000,
    0b00001000,
    0b00110110,
    0b01000001,
    0b00000000,

    0b00000000,
    0b00000000,
    0b01111111,
    0b00000000,
    0b00000000,

    0b00000000,
    0b01000001,
    0b00110110,
    0b00001000,
    0b00000000,

    0b00000111,
    0b00000101,
    0b00000111,
    0b00000000,
    0b00000000,

    0b00000010,
    0b00000100,
    0b00001000,
    0b00010000,
    0b00100000,

    /* */

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

} ;


#define  SPI_GPIO_ADDR  0x00


#define GLCD_CS1  0x01
#define GLCD_CS2  0x02
//#define GLCD_E  0x04
#define GLCD_RW  0x04
#define GLCD_DI  0x08
#define GLCD_BL  0x80

#define GLCD_DI_BIT  0x80

uint8_t
    GlcdPlus::dataGet(uint8_t bank)
{
    uint8_t data, newCtrl ;

    if (ioOut) {
        gpio.write(IODIRA, 0xff) ; // s͂
        ioOut = 0 ;
    }

    newCtrl = GLCD_RW ;
    newCtrl |= (bank & 0x01) ? GLCD_CS1:0;
    newCtrl |= (bank & 0x02) ? GLCD_CS2:0;
    newCtrl |= (bank & GLCD_DI_BIT) ?  GLCD_DI:0;
    newCtrl |= (swBackLight) ? GLCD_BL:0;

    if (ctrl != newCtrl) {
        ctrl = newCtrl ;
        gpio.write(GPIOB, ctrl) ;
//    delayMicroseconds(1) ;
    }

    digitalWrite(pinE, 1);
//    delayMicroseconds(1) ;
    digitalWrite(pinE, 0);
//    delayMicroseconds(1) ;

    digitalWrite(pinE, 1);
    data = gpio.read(GPIOA) ;
    digitalWrite(pinE, 0);

    return data ;
}

void
    GlcdPlus::dataSet(uint8_t bank, uint8_t data)
{
    uint8_t newCtrl ;

    //  ctrl = GLCD_RW ;
    newCtrl = (bank & 0x01) ? GLCD_CS1:0;
    newCtrl |= (bank & 0x02) ? GLCD_CS2:0;
    newCtrl |= (bank & GLCD_DI_BIT) ?  GLCD_DI:0;
    newCtrl |= (swBackLight) ? GLCD_BL:0;

    if (ctrl != newCtrl) {
        ctrl = newCtrl ;
        gpio.write(GPIOB, ctrl) ;
//    delayMicroseconds(1) ;
    }

    if (!ioOut) {
        gpio.write(IODIRA, 0x00) ; // so͂
        ioOut = 1 ;
    }

    digitalWrite(pinE, 1);
    gpio.write(GPIOA, data) ;
//    delayMicroseconds(1) ;
    digitalWrite(pinE, 0);
}

void
    GlcdPlus::dataPut(uint8_t bank)
{
    uint8_t ctrl ;

    //  ctrl = GLCD_RW ;
    ctrl = (bank & 0x01) ? GLCD_CS1:0;
    ctrl |= (bank & 0x02) ? GLCD_CS2:0;
    ctrl |= (bank & GLCD_DI_BIT) ?  GLCD_DI:0;
    ctrl |= (swBackLight) ? GLCD_BL:0;

    gpio.write(GPIOB, ctrl) ;
}

inline uint8_t
    GlcdPlus::dataRead(uint8_t bank)
{
    return dataGet(bank | GLCD_DI_BIT) ;
}

inline void
    GlcdPlus::dataWrite(uint8_t bank, uint8_t data)
{
    dataSet(bank | GLCD_DI_BIT, data) ;
}

inline uint8_t
    GlcdPlus::statusRead(uint8_t bank)
{
    return dataGet(bank) ;
}

inline void
    GlcdPlus::cmdWrite(uint8_t bank, uint8_t data)
{
    dataSet(bank, data) ;
}

GlcdPlus::GlcdPlus() : gpio(9, SPI_GPIO_ADDR), pinE(8)
{
    pinMode(pinE, OUTPUT);      // sets the digital pin as output
    digitalWrite(pinE, LOW);

    gpio.write(IODIRA, 0x00) ;	// f[^
    ioOut = 1 ;
    gpio.write(IODIRB, 0x70) ;	// AXCb`
    gpio.write(GPPUB, 0x70) ;

    ctrl = 0x00 ;
    gpio.write(GPIOB, ctrl) ;

    __scrollTop = 0 ;
    __scrollBottom = 7 ;

    cmdWrite(GLCD_SELECT_ALL, GLCD_DISPLAY_ON) ;
    //  delay(10) ;

    cmdWrite(GLCD_SELECT_ALL, GLCD_SET_ADDRESS | 0) ;
    //  delay(10) ;
    cmdWrite(GLCD_SELECT_ALL, GLCD_PAGE_ADDRESS_SET | 0) ;
    //  delay(10) ;

    cmdWrite(GLCD_SELECT_ALL, GLCD_DISPLAY_START_LINE) ;
    //  delay(10) ;
    /*
    for (int i = 0; i < 50; i++) {
    dataWrite(1, 0b01011101) ;
    }
    */
}

void
    GlcdPlus::clear(uint8_t cs)
{
    uint8_t p, i ;

    cmdWrite(cs, GLCD_DISPLAY_START_LINE) ;

    for (p = 0; p < 8; p++) {
        cmdWrite(cs, GLCD_PAGE_ADDRESS_SET | p) ;

        for (i = 0; i < 64; i++)
            dataWrite(cs, 0b00000000) ;
    }

    cmdWrite(cs, GLCD_SET_ADDRESS | 0) ;
    cmdWrite(cs, GLCD_PAGE_ADDRESS_SET | 0) ;
}


void
    GlcdPlus::pset(uint8_t x, uint8_t y, uint8_t pixel)
{
    uint8_t p, col, cs ;
    uint8_t d ;

    if (x < 64) {
        cs = GLCD_SELECT_LEFT ;
        col = x ;
    } 
    else {
        cs = GLCD_SELECT_RIGHT ;
        col = x - 64 ;
    }

    d = (1 << (7 - (y % 8))) ;
    //  d = 32 ;
    p = 7 - (y / 8) ;

    cmdWrite(cs, GLCD_PAGE_ADDRESS_SET + p) ;
    cmdWrite(cs, GLCD_SET_ADDRESS + col) ;

    if (pixel == GLCD_PIXEL_CLEAR)
        d = ~d & dataRead(cs) ;
    else if (pixel == GLCD_PIXEL_XOR)
        d ^= dataRead(cs) ;
    else
        d |= dataRead(cs) ;

    cmdWrite(cs, GLCD_SET_ADDRESS + col) ;
    dataWrite(cs, d) ;
}


/***********************************************************
̕`
***********************************************************/

#define BIT(x)	(1 << (x))

void
    GlcdPlus::drawLine(uint8_t x, uint8_t y, uint8_t x1, uint8_t y1, uint8_t pixel)
{
    int dx, dy, adx, ady, dd, da ;
    uint8_t tmp ;
    uint8_t pg, col, cs ;
    uint8_t d0, d ;

    if (x > GLCD_X_MAX)
        return ;

    if (y > GLCD_Y_MAX)
        return ;

    if (x1 > GLCD_X_MAX)
        return ;

    if (y1 > GLCD_Y_MAX)
        return ;

    dx = x - x1 ;
    dy = y - y1 ;

    if (dx == 0) {
        /* c̒ */
        if (y > y1) {
            tmp = y1 ;
            y1 = y ;
            y = tmp ;
        }

        if (x < 64) {
            cs = GLCD_SELECT_LEFT ;
            col = x ;
        } else {
            cs = GLCD_SELECT_RIGHT ;
            col = x - 64 ;
        }

        for ( ; y <= y1; y++) {
            d = BIT(7 - (y & 0b111)) ;
            pg = 7 - (y >> 3) ;

            cmdWrite(cs, GLCD_PAGE_ADDRESS_SET | pg) ;
            cmdWrite(cs, GLCD_SET_ADDRESS | col) ;

            if (pixel == GLCD_PIXEL_CLEAR)
                d = ~d & dataRead(cs) ;
            else if (pixel == GLCD_PIXEL_XOR)
                d ^= dataRead(cs) ;
            else
                d |= dataRead(cs) ;

            cmdWrite(cs, GLCD_SET_ADDRESS | col) ;
            dataWrite(cs, d) ;
        }	
    } else if (dy == 0) {
        /* ̒ */
        if (x > x1) {
            tmp = x1 ;
            x1 = x ;
            x = tmp ;
        }

        d0 = BIT(7 - (y & 0b111)) ;
        pg = 7 - (y >> 3) ;
        cmdWrite(GLCD_SELECT_ALL, GLCD_PAGE_ADDRESS_SET | pg) ;

        for ( ; x <= x1; x++) {
            if (x < 64) {
                cs = GLCD_SELECT_LEFT ;
                col = x ;
            } else {
                cs = GLCD_SELECT_RIGHT ;
                col = x - 64 ;
            }

            cmdWrite(cs, GLCD_SET_ADDRESS | col) ;

            if (pixel == GLCD_PIXEL_CLEAR)
                d = ~d0 & dataRead(cs) ;
            else if (pixel == GLCD_PIXEL_XOR)
                d = d0 ^ dataRead(cs) ;
            else
                d = d0 | dataRead(cs) ;

            cmdWrite(cs, GLCD_SET_ADDRESS | col) ;
            dataWrite(cs, d) ;
        }	
    } else {
        /* ΂߂̒ */
        adx = dx > 0 ? dx : -dx ;
        ady = dy > 0 ? dy : -dy ;
        if (adx > ady) {
            dd = dy << 8 ;
            dd /= dx ;

            if (x > x1) {
                tmp = x1 ;
                x1 = x ;
                x = tmp ;
                dd = -dd ;
            }

            da = y << 8 ;
            for ( ; x <= x1; x++) {
                pset(x, da >> 8, pixel) ;
                da += dd ;
            }	
        } else {
            dd = dx << 8 ;
            dd /= dy ;

            if (y > y1) {
                tmp = y1 ;
                y1 = y ;
                y = tmp ;
                dd = -dd ;
            }

            da = x << 8 ;
            for ( ; y <= y1; y++) {
                pset(da >> 8, y, pixel) ;
                da += dd ;
            }	
        }	
    }	
}


/***********************************************************
	`̕`
***********************************************************/

void
    GlcdPlus::drawRectangle(uint8_t x, uint8_t y, uint8_t x1, uint8_t y1, uint8_t pixel)
{
    if (x > GLCD_X_MAX)
        return ;

    if (y > GLCD_Y_MAX)
        return ;

    if (x1 > GLCD_X_MAX)
        return ;

    if (y1 > GLCD_Y_MAX)
        return ;

    drawLine(x, y, x, y1, pixel) ;
    drawLine(x1, y, x1, y1, pixel) ;
    drawLine(x, y, x1, y, pixel) ;
    drawLine(x, y1, x1, y1, pixel) ;
}	


/***********************************************************
	obNCg
***********************************************************/

void
    GlcdPlus::backLight(uint8_t sw)
{
    swBackLight = sw;
    dataPut(0) ;
}


/***********************************************************
	J[\ʒu
***********************************************************/

void
    GlcdPlus::setCursor(uint8_t col, uint8_t line)
{
    _line = line ;
    _col= col ;
}


/***********************************************************
	\̃XN[
***********************************************************/


void
    GlcdPlus::setScroll(uint8_t top, uint8_t bottom)
{
    if (top >= bottom)
        return ;

    if ((top < 0) || (bottom < 0))
        return ;

    if ((top > 7) || (bottom > 7))
        return ;

    __scrollTop = top ;
    __scrollBottom = bottom ;

    _line = top ;
}


void
    GlcdPlus::scroll(uint8_t top, uint8_t bottom)
{
    uint8_t i ;
    uint8_t p, cs ;
    uint8_t buff[64] ;

    cmdWrite(GLCD_SELECT_ALL, GLCD_DISPLAY_START_LINE | 0) ;

    for (cs = 1; cs < 3; cs++) {
        for (p = top; p < bottom; p++) {
            /* ̍s̃f[^ǂݏo */
            cmdWrite(cs, GLCD_PAGE_ADDRESS_SET | (p + 1)) ;

            for (i = 0; i < 64; i++) {
                cmdWrite(cs, GLCD_SET_ADDRESS | i) ;
                buff[i] = dataRead(cs) ;
            }	

            /* ̍sɏ */
            cmdWrite(cs, GLCD_PAGE_ADDRESS_SET | p) ;
            cmdWrite(cs, GLCD_SET_ADDRESS | 0) ;

            for (i = 0; i < 64; i++)
                dataWrite(cs, buff[i]) ;
        }
    }

    /* ŉs̃NA */
    cmdWrite(GLCD_SELECT_ALL, GLCD_PAGE_ADDRESS_SET | bottom) ;
    cmdWrite(GLCD_SELECT_ALL, GLCD_SET_ADDRESS | 0) ;

    for (i = 0; i < 64; i++)
        dataWrite(GLCD_SELECT_ALL, 0b00000000) ;	

    cmdWrite(GLCD_SELECT_ALL, GLCD_SET_ADDRESS | 0) ;
    cmdWrite(GLCD_SELECT_ALL, GLCD_PAGE_ADDRESS_SET | 0) ;
}


void
    GlcdPlus::write(uint8_t c)
{
    uint8_t col, cs, i ;
    uint8_t d ;

    if (c == '\r') {	// A
        _col = 0 ;
        return ;
    } else if (c == '\n') {	// s
        if (++_line > __scrollBottom) {
            _line = __scrollBottom ;
            scroll(__scrollTop, __scrollBottom) ;
        }	
        return ;
    } else if (c < 32)
        return ;

    if ((col = _col++ * 6) < 64) {
        cs = 1 ;
    } 
    else {
        cs = 2 ;
        col -= 64 ;
    }

    cmdWrite(cs, GLCD_PAGE_ADDRESS_SET + _line) ;
    cmdWrite(cs, GLCD_SET_ADDRESS + col) ;

    c -= 32 ;
    for (i = 0; i < 5; i++, col++) {
        if (col == 64) {
            cs = 2 ;
            cmdWrite(cs, GLCD_PAGE_ADDRESS_SET + _line) ;
            cmdWrite(cs, GLCD_SET_ADDRESS) ;
        }
        //    d = pgm_read_uint8_t(&FONT[c*5 + i]) ;
        d = pgm_read_byte_near(&FONT[c*5 + i]) ;
        dataWrite(cs, d) ;
    }

    dataWrite(cs, 0b00000000) ; // ƕƂ̊Ԃ̋

    if (_col >= 21) {
        _col = 0 ;
        if (++_line > __scrollBottom) {
            _line = __scrollBottom ;
            scroll(__scrollTop, __scrollBottom) ;
        }	
    }
}

uint8_t
    GlcdPlus::sw1(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return (v & 0x10) ? LOW : HIGH;
}

uint8_t
    GlcdPlus::sw2(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return (v & 0x20) ? LOW : HIGH;
}

uint8_t
    GlcdPlus::sw3(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return (v & 0x40) ? LOW : HIGH;
}

uint8_t
    GlcdPlus::swAll(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return ~v >> 4;
}
